#!/usr/bin/perl
use strict;

use C4::Koha;
use C4::Context;

use PDF::Reuse;
use MARC::Record;
use MARC::File::XML;
use CGI;
use Roman;

use constant Y_TOP    => 730;    ## Top of first card
use constant Y_OFFSET => 220;    ## Distance between the tops of each card

use constant X_LEFT     => 130;            ## The left edge of the card
use constant X_INDENTED => X_LEFT + 50;    ## The left edge of the text
use constant X_CENTER   => 325;            ## The horizontal center of the card
use constant X_RIGHT    => 455;            ## The right-hand edge of the card

use constant INDENT => 15;                 ## The distance an indent represents

use constant FH => 8;                     ## Font Height

use constant MAX_LINES     => 21;          ## Total lines per card
use constant STARTING_LINE => 2;

use constant MAX_STRING_WIDTH     => 65;    ## Max number of characters per line
use constant MAX_HEADING_WIDTH	  => 57;
use constant MAX_CALLNUMBER_WIDTH => 6;    ## The max characters per line for the callnumber

my $dbh = C4::Context->dbh;

my $input        = new CGI;
my $barcodeStart = $input->param('barcodeStart');
my $barcodeEnd   = $input->param('barcodeEnd');

if ( !$barcodeEnd ) {
    $barcodeEnd = $barcodeStart;
}

my $printMainEntry = $input->param('printMainEntry');
my $printShelf     = $input->param('printShelf');
my $printAuthor    = $input->param('printAuthor');
my $printSubject   = $input->param('printSubject');
my $printTitle     = $input->param('printTitle');

our @cardTracings;
our $_marc;
our $_item;
our $_sortingLine;
our $_currentCard_number;

## Outputing a PDF file directly to the browser
prInitVars();
$| = 1;
print STDOUT "Content-Disposition: attachment;filename=\"cards.pdf\"\n";
print STDOUT "Content-Type: application/pdf \r\n\r\n";
prFile();

#prFont('Courier');
prFontSize(FH);

my $currentCard = 1;
for ( my $barcode = $barcodeStart ; $barcode <= $barcodeEnd ; $barcode++ ) {
    $_currentCard_number = 1;

    my $sth = $dbh->prepare(
        "SELECT * FROM items, biblio, biblioitems
                          WHERE items.biblionumber = biblio.biblionumber
                          AND items.biblionumber = biblioitems.biblionumber
                          AND items.barcode = ?"
    );
    $sth->execute($barcode);

    $_item = $sth->fetchrow_hashref;

    $_marc = MARC::Record->new_from_xml( $_item->{'marcxml'} );

    ## Build the Tracings
    my $cardsList;
    my @tracings;

    my @subjects = getSubjects($_item);
    my $i        = 1;
    foreach my $subject (@subjects) {
#        $cardsList .= "$i. $subject  ";
        push( @tracings, "$i. $subject  " );
        $i++;
    }

    my $romanNumeral = 1;
    if ( defined( $_marc->field('100') ) ) {
#       $cardsList .= Roman( $romanNumeral++ ) . ". Title ";
        push( @tracings, Roman( $romanNumeral++ ) . ". Title " );
    }
    
    if ( defined( $_marc->field('110') ) ) {
      push( @tracings, Roman( $romanNumeral++ ) . ". Title " );
    }

    my @additionalAuthors = getAuthors($_item);
    foreach my $author (@additionalAuthors) {
#        $cardsList .= Roman( $romanNumeral++ ) . ". $author  ";
        push( @tracings, Roman( $romanNumeral++ ) . ". $author  " );
    }
    
#    @cardTracings = formatTracings( @tracings );
    @cardTracings = join( '', @tracings );
    
    ## Main Entry Card
    if ($printMainEntry) {
        $currentCard = createCard( $_item, $currentCard, $cardsList );
    }

    if ($printShelf) {
        $currentCard = createShelfCard( $_item, $currentCard, $cardsList );
    }

    if ($printTitle) {
        if ( $_item->{'author'} || $_marc->field('110') ) { ## If we have an author, we need a Title Card
            my $titleLine = getTitle($_item);

            $currentCard =
              createCard( $_item, $currentCard, $cardsList, $titleLine );
        }

        ## Additional Title
        my @_740s = $_marc->field('740');
        foreach my $_740 ( @_740s ) {
          if ( $_740->subfield("a") ) {
              $currentCard =
                createCard( $_item, $currentCard, $cardsList,
                  $_740->subfield("a") );
          }
        }
    }

    ## Need a Subject Card for each subject
    if ($printSubject) {
        foreach my $subject (@subjects) {
            $currentCard =
              createCard( $_item, $currentCard, $cardsList, uc($subject) );
        }
    }

    ## Additional Author Cards
    if ($printAuthor) {
        my @authors = $_marc->field('700');
        foreach my $author (@authors) {
            my $line;
            $line .= getSubfields( $author, 'a' );
            $line .= getSubfields( $author, 'q' );
            $line .= getSubfields( $author, 'b' );
            $line .= getSubfields( $author, 'c' );
            $line .= getSubfields( $author, 'd' );
            $line .= ' ' . getSubfields( $author, 'e' );
            $line .= getSubfields( $author, 't' );
            $currentCard = createCard( $_item, $currentCard, $cardsList, $line );
        }
        
        @authors = $_marc->field('710');
        foreach my $author (@authors) {
            my $line;
            $line .= getSubfields( $author, 'a' );
            $line .= getSubfields( $author, 'b' );
            $line .= getSubfields( $author, 'e' );
            $line .= getSubfields( $author, 't' );

            $currentCard = createCard( $_item, $currentCard, $cardsList, $line );
        }
        
    }
}

prEnd();

sub getBibData {
    my ($item) = @_;

    my @lines;

    my $line;

    $line = '';

    ## Title
    if ( $item->{'title'} ) {
        $line .= $item->{'title'};    ## Title
    }
    
    ## Medium
    if ( defined( $_marc->field('245') ) && defined( $_marc->field('245')->subfield('h') ) ) {
        $line .= ' ' . $_marc->field('245')->subfield('h');
    }  

    ## Subtitle i.e. Remainder of title
    if ( $_marc->field('245') && $_marc->field('245')->subfield("b") ) {
        $line .= ' ' . $_marc->field('245')->subfield("b");
    }

    ## Statement of responsibility, etc.
    if ( $_marc->field('245') && $_marc->field('245')->subfield("c") ) {
        $line .= " ";                 ## Separator
        $line .=
          $_marc->field('245')->subfield("c")
          ;                           ## Statement of responsibility, etc.
    }

    my @field246 = $_marc->field('246');
    foreach my $field246 ( @field246 ) {

      ## 246 i Varying form of title
      if ( $field246 && $field246->subfield("i") ) {
          $line .= " ";                 ## Separator
          $line .= $field246->subfield("i");
      }
    
      ## 246 a 
      if ( $field246 && $field246->subfield("a") ) {
          $line .= " : ";                 ## Separator
          $line .= $field246->subfield("a");
      }
    
      ## 246 b
      if ( $field246 && $field246->subfield("b") ) {
          $line .= " : ";                 ## Separator
          $line .= $field246->subfield("b");
      } else {
        $line .= ". "; ## Period after 246a if there is no 246b
      }

      ## 246 n
      if ( $field246 && $field246->subfield("n") ) {
          $line .= " .";                 ## Separator
          $line .= $field246->subfield("n");
      }

      ## 246 h
      if ( $field246 && $field246->subfield("h") ) {
          $line .= " ";                 ## Separator
          $line .= $field246->subfield("h");
      }

      ## 246 f
      if ( $field246 && $field246->subfield("f") ) {
          $line .= " ";                 ## Separator
          $line .= $field246->subfield("f");
      }

    }

    ## Edition
    if ( $_marc->field('250') && $_marc->field('250')->subfield("a") ) {
        $line .= " -- ";                               ## Separator
        $line .= $_marc->field('250')->subfield("a");
    }

    ## Publisher
    $line .= " -- ";                                              ## Separator
    $line .= $item->{'place'} . " " . $item->{'publishercode'};

    ## Copyright
    if ( $_marc->field('260') && $_marc->field('260')->subfield("c") ) {
        $line .= ' ' . $_marc->field('260')->subfield("c");
    }

    push( @lines, $line );
    $line = '';

    ## Pages i.e. Physical Extent
    if ( $_marc->field('300') && $_marc->field('300')->subfield("a") ) {
        $line .= $_marc->field('300')->subfield("a") . " ";
    }
    if ( $_marc->field('300') && $_marc->field('300')->subfield("b") ) {
        $line .= $_marc->field('300')->subfield("b") . " ";
    }
    if ( $_marc->field('300') && $_marc->field('300')->subfield("c") ) {
        $line .= $_marc->field('300')->subfield("c") . " ";
    }
    if ( $_marc->field('300') && $_marc->field('300')->subfield("e") ) {
        $line .= $_marc->field('300')->subfield("e") . " ";
    }
    
    ## 310 Current Publication Frequency
    if ( $_marc->field('310') && $_marc->field('310')->subfield("a") ) {
      $line .= $_marc->field('310')->subfield("a");
    }
    if ( $_marc->field('310') && $_marc->field('310')->subfield("b") ) {
      $line .= ', '; ## 310a ends with comma
      $line .= $_marc->field('310')->subfield("b");
    }
#    unless ( 
#      ( substr( $line, -1 ) eq '.' )
#      ||
#      ( substr( $line, -1 ) eq '!' )
#      ||
#      ( substr( $line, -1 ) eq '?' )
#    ) {
#      $line .= '.';
#    }
    $line .= ' ';
    

    ## Series
    if ( $_marc->field('490') && $_marc->field('490')->subfield('a') ) {
        $line .= "(" . $_marc->field('490')->subfield('a');

        if ( $_marc->field('490') && $_marc->field('490')->subfield('v') ) {
            $line .= ' ' . $_marc->field('490')->subfield('v');
        }
        $line .= ")";
    }

    #    if ( $item->{'seriestitle'} ) {
    #        $line .= "( " . $item->{'seriestitle'} . " ) ";
    #    }

    push( @lines, $line );

    push( @lines, "\n" );

    ## 500 General Note
    $line = '';
    my $numberOfFields = 0;
    my @fields         = $_marc->field('500');
    foreach my $field (@fields) {
        if ( ++$numberOfFields > 1 ) { $line .= '--'; }

        my $numberOfaFields = 0;
        my @a_fields        = $field->subfield('a');
        foreach my $a_field (@a_fields) {
            if ( ++$numberOfaFields > 1 ) { $line .= '--'; }
            $line .= defined($a_field) ? " " . $a_field : '';
        }

    }
    if (@fields) {
#        $line .= '.';
        push( @lines, $line );
    }

    ## 504 Bibliography, etc.. Note
    $line = '';
    my $numberOfFields = 0;
    my @fields         = $_marc->field('504');
    foreach my $field (@fields) {
        if ( ++$numberOfFields > 1 ) { $line .= '--'; }
        $line .= defined( $field->subfield('a') ) ? $field->subfield('a') : '';
    }
    if (@fields) {
#        $line .= '.';
        push( @lines, $line );
    }

    ## 505 -FORMATTED CONTENTS NOTE
    $line = '';
    my $numberOfFields = 0;
    my @fields         = $_marc->field('505');
    foreach my $field (@fields) {
        if ( ++$numberOfFields > 1 ) { $line .= '--'; }
        $line .= defined( $field->subfield('a') ) ? $field->subfield('a') : '';
        $line .=
          defined( $field->subfield('g') ) ? " " . $field->subfield('g') : '';

        my $numberOftFields = 0;

        my @t_fields        = $field->subfield('t');
        my @r_fields        = $field->subfield('r');
        ## The field arrays are pulled in reverse order, fix it.
#        my @t_fields = reverse( @t_fields );
#        my @r_fields = reverse( @r_fields );

        while ( @t_fields && @r_fields ) {
            my $t_field = pop( @t_fields );
            my $r_field = pop( @r_fields );
            if ( ++$numberOftFields > 1 ) { $line .= ' --'; }
            $line .= defined($t_field) ? " " . $t_field : '';
            $line .= defined($r_field) ? " / " . $r_field : '';
        }

        my $numberOfrFields = 0;
        foreach my $t_field (@t_fields) {
            if ( ++$numberOftFields > 1 ) { $line .= ' --'; }
            $line .= defined($t_field) ? " " . $t_field : '';
        }

        my $numberOfrFields = 0;
        foreach my $r_field (@r_fields) {
            if ( ++$numberOfrFields > 1 ) { $line .= ' --'; }
            $line .= defined($r_field) ? " " . $r_field : '';
        }
        
        $line .= '.';
    }
    if (@fields) {
#        $line .= '.';
        push( @lines, $line );
    }

    ## 511 Participant or Performer Note
    $line = '';
    my @fields = $_marc->field('511');
    foreach my $field ( @fields ) {
      $line .= ' ' .  $field->subfield('a');
    }

    ## 520 Summary, Etc. Note
    $line = '';
    my $numberOfFields = 0;
    my @fields         = $_marc->field('520');
    foreach my $field (@fields) {
        if ( ++$numberOfFields > 1 ) { $line .= '--'; }
        $line .= defined( $field->subfield('a') ) ? $field->subfield('a') : '';
    }
    if (@fields) {
#        $line .= '.';
        push( @lines, $line );
    }

    ## 521 Target Audience Note
    $line = '';
    $numberOfFields = 0;
    @fields         = $_marc->field('521');
    foreach my $field (@fields) {
      my @subfield_a = $field->subfield('a');
      foreach my $s (@subfield_a) {
        if ( ++$numberOfFields > 1 ) { $line .= ' -- '; }
        $line .= $s;
      }
      
      if ( defined( $field->subfield('b') ) ) {
        $line .= ' ' . $field->subfield('b') . ' ';
      }
    }
    if (@fields) {
#        $line .= '.';
        push( @lines, $line );
    }
    

    ## 533 Reproduction Note
    $line = '';
    $numberOfFields = 0;
    @fields         = $_marc->field('533');
    foreach my $field (@fields) {
      $line .= $field->subfield('a') . '. ';
    
      my @subfield_b = $field->subfield('b');
      foreach my $s (@subfield_b) {
        if ( ++$numberOfFields > 1 ) { $line .= ' : '; }
        $line .= $s;
      }
      
      $line .= ' : ';
      
      my @subfield_c = $field->subfield('c');
      foreach my $s (@subfield_c) {
        if ( ++$numberOfFields > 1 ) { $line .= ' : '; }
        $line .= $s;
      }

      if ( defined( $field->subfield('d') ) ) {
        $line .= ', ';    
        $line .= $field->subfield('d') . '. ';
      } else {
        $line .= '. '; ## If there is a d, there is a comma after c, otherwise it's a period
      }        

      if ( defined( $field->subfield('e') ) ) {
        $line .= $field->subfield('e') . '. ';
      }

      my @subfield_f = $field->subfield('f');
      foreach my $s (@subfield_f) {
        $line .= "($s). ";
      }
    
      my @subfield_m = $field->subfield('m');
      foreach my $s (@subfield_m) {
        $line .= " $s ";
      }

      my @subfield_n = $field->subfield('n');
      foreach my $s (@subfield_n) {
        $line .= " $s ";
      }

      if ( defined( $field->subfield('3') ) ) {
        $line .= $field->subfield('3') . ". ";
      }

    }
    if (@fields) {
#        $line .= '.';
        push( @lines, $line );
    }


    ## 586 Awards Note
    $line = '';
    my $numberOfFields = 0;
    my @fields         = $_marc->field('586');
    foreach my $field (@fields) {
        if ( ++$numberOfFields > 1 ) { $line .= '--'; }
        $line .= defined( $field->subfield('a') ) ? $field->subfield('a') : '';
    }
    if (@fields) {
#        $line .= '.';
        push( @lines, $line );
    }

    return @lines;
}

sub getShelfData {
    my ($item) = @_;

    my @lines;
    my $line;

    if ( $item->{'price'} ) {
        $line .= '$' . $item->{'price'} . ' ';
    }
    $line .= 'BC#' . $item->{'barcode'};
    push( @lines, $line );

    my $line = $item->{'dateaccessioned'};
    push( @lines, $line );

    return @lines;
}

sub createCard {
    my ( $item, $cardNumber, $cardsList, $header ) = @_;

    $_currentCard_number = 1;
    
    if ( $item->{'author'} ) {
      $_sortingLine = getAuthor();
    } else {
      $_sortingLine = getTitle( 1 );
    }

    my $yo = Y_OFFSET * ( $cardNumber - 1 );

    if ($header) {
        printHeader( $cardNumber, $header );
    }

    my $line = STARTING_LINE;

    printCallnumber( $item, $cardNumber );

    my $hasAuthor = 0;
    
    ## Main Author
    if ( $item->{'author'} ) {
        $hasAuthor = 1;
        
        my $authorLine = getAuthor( $item );
        ( $cardNumber, $line ) = printLine( $authorLine, $cardNumber, $line, my $indent = 0, $item, $header );
        
    } elsif ( $_marc->field('110') && $_marc->field('110')->subfield('a') ) {

        $hasAuthor = 1;    
        my $authorLine = $_marc->field('110')->subfield('a') . ". ";
        
        if ( $_marc->field('110') && $_marc->field('110')->subfield('b') ) {
        
          my @subfields = $_marc->field('110')->subfield('b');
        
          foreach my $subfield ( @subfields ) {
            $authorLine .= "$subfield. ";
          }
          
        }
    
        if ( $_marc->field('110') && $_marc->field('110')->subfield('t') ) {
          $authorLine .= ". " . $_marc->field('110')->subfield('t');
        }
                                                                                                            
        ( $cardNumber, $line ) = printLine( $authorLine, $cardNumber, $line, my $indent = 0, $item, $header );
                                                                                                                                           
    } 

    my $indentStyle = 'standard';
    if ( ! $hasAuthor ) {
        $indentStyle = 'hanging';
    }

    my (@bibLines) = getBibData($item);
    ( $cardNumber, $line ) =
      printLines( $cardNumber, $line, $indentStyle, $item, $header, @bibLines );

    $line++;

    my $lccn;
    if ( $_marc->field('010') && $_marc->field('010')->subfield('a') ) {
        $lccn = $_marc->field('010')->subfield('a');
    }
    if ($lccn) {
        ( $cardNumber, $line ) = printLine(
            "LCCN: " . $lccn,
            $cardNumber, $line, my $indent = 1,
            $item, $header
        );
    }

    my @isbns     = $_marc->field('020');
    my $isbnCount = 0;
    my $isbnLine  = 'ISBN: ';
    foreach my $isbn (@isbns) {
        if ( ++$isbnCount > 1 ) {
            $isbnLine .= " | ";
        }
        $isbnLine .= $isbn->subfield('a');
    }
    ( $cardNumber, $line ) =
      printLine( $isbnLine, $cardNumber, $line, my $indent = 1, $item,
        $header );

    $line++;
    ( $cardNumber, $line ) =
      printLines( $cardNumber, $line, $indentStyle = 'standard',
        $item, $header, @cardTracings );

#    my $indent = 1;
#    foreach my $ct ( @cardTracings ) {
#      ( $cardNumber, $line ) = printLine( $ct, $line, $indent, $item, $header );
#      $indent = 0;
#    }

    return getNextCardNum($cardNumber);
}

sub createShelfCard {
    my ( $item, $cardNumber, $cardsList, $header ) = @_;

    $_currentCard_number = 1;

    my $yo = Y_OFFSET * ( $cardNumber - 1 );

    if ($header) {
        printHeader( $cardNumber, $header );
    }

    my $line = STARTING_LINE;

    printCallnumber( $item, $cardNumber );

    ## Main Author
#    warn "AUTHOR: " . $item->{'author'};
    if ( $item->{'author'} ) {

        my $authorLine = getAuthor();
        if ( $_marc->field('100')->subfield('c') ) {
            $authorLine .= ' ' . $_marc->field('100')->subfield('c');
        }

        ( $cardNumber, $line ) =
          printLine( $authorLine, $cardNumber, $line, my $indent = 0,
            $item, $header );
    } elsif ( $_marc->field('110') && $_marc->field('110')->subfield('a') ) {
        my $authorLine = $_marc->field('110')->subfield('a');
        if ( $_marc->field('110') && $_marc->field('110')->subfield('b') ) {
            my @subfields = $_marc->field('110')->subfield('b');
            foreach my $subfield ( @subfields ) {
                $authorLine .= ". $subfield";
            }
        }
        if ( $_marc->field('110') && $_marc->field('110')->subfield('t') ) {
            $authorLine .= ". " . $_marc->field('110')->subfield('t');
        }
        
        ( $cardNumber, $line ) =
          printLine( $authorLine, $cardNumber, $line, my $indent = 0,
          $item, $header );
                              

    } else {
#      warn "NO AUTHOR FOUND";
    }

    my $indentStyle = 'standard';
    if ( !$item->{'author'} ) {
        $indentStyle = 'hanging';
    }

    my (@bibLines) = getBibData($item);
    ( $cardNumber, $line ) =
      printLines( $cardNumber, $line, $indentStyle, $item, $header, @bibLines );

    my @shelfData = getShelfData($item);
    ( $cardNumber, $line ) =
      printLines( $cardNumber, $line, $indentStyle, $item, $header,
        @shelfData );

    my $lccn;
    if ( $_marc->field('010') && $_marc->field('010')->subfield('a') ) {
        $lccn = $_marc->field('010')->subfield('a');
    }
    if ($lccn) {
        ( $cardNumber, $line ) = printLine(
            "LCCN: " . $lccn,
            $cardNumber, $line, my $indent = 1,
            $item, $header
        );
    }

    my @isbns     = $_marc->field('020');
    my $isbnCount = 0;
    my $isbnLine  = 'ISBN: ';
    foreach my $isbn (@isbns) {
        if ( ++$isbnCount > 1 ) {
            $isbnLine .= " | ";
        }
        $isbnLine .= $isbn->subfield('a');
    }
    ( $cardNumber, $line ) =
      printLine( $isbnLine, $cardNumber, $line, my $indent = 1, $item,
        $header );

#    $line++;
#    my @cardsList;
#    push( @cardsList, $cardsList );
#    ( $cardNumber, $line ) =
#      printLines( $cardNumber, $line, $indentStyle = 'standard',
#        $item, $header, @cardsList );

    $line++;
    ( $cardNumber, $line ) =
      printLines( $cardNumber, $line, $indentStyle = 'standard',
        $item, $header, @cardTracings );

    return getNextCardNum($cardNumber);
}

sub printLines {
    my ( $cardNumber, $line, $indentStyle, $item, $header, @text ) = @_;

    ## Prime the special indenting for first line
    my $x_indents = 0;
    if ( $indentStyle eq 'standard' ) {
        $x_indents = 1;
    }
    if ( $indentStyle eq 'standard_' ) { ## No re-indent after first line
        $x_indents = 1;
    }

    foreach my $textLine (@text) {
#warn "PrintLines: XI: $x_indents :: TEXT: $textLine ";
        my @words = split( / /, $textLine );
        my $currentLine;

        my $words = @words;
        while ($words) {
            my $currentWord = shift(@words);
            unshift( @words, $currentWord );

            while (
                length($currentLine) <=
                MAX_STRING_WIDTH - ( INDENT * $x_indents ) -
                length($currentWord) )
            {
                $currentLine .= shift(@words) . " ";
                $words = @words;
            }
            ( $cardNumber, $line ) =
              printLine( $currentLine, $cardNumber, $line, $x_indents, $item,
                $header );

            $currentLine = '';
            ## After first line, all indenting is the same
            if ( $indentStyle eq 'standard' ) {
                $x_indents = 0;
            }
            elsif ( $indentStyle eq 'hanging' ) {
                $x_indents = 1;
            }

        }

        ## Special indenting for first line
        $x_indents = 0;
        if ( $indentStyle eq 'standard' ) {
            $x_indents = 1;
        }

    }

    return ( $cardNumber, $line );
}

sub printLine {
    my ( $text, $cardNumber, $line, $x_indents, $item, $header ) = @_;
#warn "PRINTLINE ACCEPTING ( $cardNumber, $line )";
#warn "PRINTLINE TEXT: $text";
    if ( $line < MAX_LINES ) {
        prText( X_INDENTED + ( INDENT * $x_indents ),
            Y_TOP - Y_OFFSET * ( $cardNumber - 1 ) - FH * $line, $text );
    }
    else { ## Has run over onto another card
        prText(
            X_INDENTED,
            Y_TOP - Y_OFFSET * ( $cardNumber - 1 ) - FH * $line,
            'CONTINUED ON NEXT CARD'
        );
        $cardNumber = getNextCardNum( $cardNumber, $item, $header );
        $line = STARTING_LINE;
        
        ## On 'continued cards', print the number in the upper right
        $_currentCard_number++;
        printCardNumber( $cardNumber);
        
        ## On 'continued cards' we print the sorting line at the top of each one.
        ( $cardNumber, $line ) = printLine( $_sortingLine, $cardNumber, $line, 0, $item, $header );
        $line++; ## Add a space before printing the rest of the data.
        
        ( $cardNumber, $line ) =
          printLine( $text, $cardNumber, $line, $x_indents, $item, $header );
    }
    
    $line++;
#warn "PRINTLINE RETURNING ( $cardNumber, $line )";
    return ( $cardNumber, $line );
}

sub printHeader {
    my ( $cardNumber, $header ) = @_;

    chomp($header);
    if ( $header =~ /\/$/ ) {    ## If line ends with a '/'
        chop($header);
    }
#    if ( $header =~ /\.$/ ) {    ## If line ends with a '.'
#        chop($header);
#    }

#warn "createcards->printHeader: $header Length: " . length( $header ) . " MAX_HEADING_WIDTH: " . MAX_HEADING_WIDTH;    
    my @header;
    if ( length( $header ) > MAX_HEADING_WIDTH ) {
      my @values = split(' ', $header);
      my $current_line;
      foreach my $val ( @values ) {
        if ( length($current_line) + length( $val ) > MAX_HEADING_WIDTH ) {
          push( @header, $current_line );
          $current_line = $val;
        } else {
          $current_line .= " " . $val;
        }
        
#        warn "current_line is $current_line";
      }
      if ( length($current_line ) ) { push( @header, $current_line ); }
    } else {
      push( @header, $header );
    }

    my $i = 0;
    foreach my $h ( @header ) {
#      warn "Printing Header Line: $h";
      prText( X_INDENTED + ( INDENT * 1 ), Y_TOP - Y_OFFSET * ( $cardNumber - 1 ) - ( FH * $i++ ), $h );
    }
}

sub printCallnumber {
    my ( $item, $cardNumber ) = @_;
    my @lines = split( / /, $item->{'itemcallnumber'} );

    my $l = STARTING_LINE;
    foreach my $line (@lines) {
        my @lines;
        while ( length($line) > 0 ) {
            push( @lines, substr( $line, 0, MAX_CALLNUMBER_WIDTH, '' ) );
        }

        foreach my $line (@lines) {
            prText( X_LEFT, Y_TOP - Y_OFFSET * ( $cardNumber - 1 ) - FH * $l++,
                $line );
        }
    }
}

## If we are continuing an entry that needs more than one card, we must write the
## callnumber and header again, that is why they are optional arguments.
sub getNextCardNum {
    my ( $cardNumber, $item, $header ) = @_;

    $cardNumber++;

    if ( $cardNumber > 3 ) {
        $cardNumber = 1;
        prPage();
    }

    if ($item) {
        printCallnumber( $item, $cardNumber );
    }

    if ($header) {
        printHeader( $cardNumber, $header );
    }

    return $cardNumber;
}

sub getSubjects {
    my ($item) = @_;

    my @subjects;

    ## 600 Subject Added Entry -- Personal Name
    my @fields = $_marc->field('600');

    foreach my $field (@fields) {
        my $subject;
        $subject .=
          defined( $field->subfield('a') ) ? $field->subfield('a') : '';
          warn "CP " . $field->subfield('a');
        $subject .=
          defined( $field->subfield('q') ) ? " " . $field->subfield('q') : '';
        $subject .=
          defined( $field->subfield('b') ) ? " " . $field->subfield('b') : '';
        $subject .=
          defined( $field->subfield('c') ) ? " " . $field->subfield('c') : '';
        $subject .=
          defined( $field->subfield('d') ) ? " " . $field->subfield('d') : '';
        $subject .=
          defined( $field->subfield('t') ) ? ". " . $field->subfield('t') : '';
        if ( defined( $field->subfield('x') ) ) {
          my @subfields = $field->subfield('x');
          foreach my $s ( @subfields ) {
            $subject .= " -- " . $s;
            warn "CP $s";
          }
        }
        if ( defined( $field->subfield('y') ) ) {
          my @subfields = $field->subfield('y');
          foreach my $s ( @subfields ) {
            $subject .= " -- " . $s;
          }
        }
        if ( defined( $field->subfield('z') ) ) {
          my @subfields = $field->subfield('z');
          foreach my $s ( @subfields ) {
            $subject .= " -- " . $s;
          }
        }
        if ( defined( $field->subfield('v') ) ) {
          my @subfields = $field->subfield('v');
          foreach my $s ( @subfields ) {
            $subject .= " -- " . $s;
          }
        }
        push( @subjects, $subject );
    }

    ## 610 Subject Added Entry -- Corporate Name
    my @fields = $_marc->field('610');
    my $firstField = 1;
    foreach my $field (@fields) {
        my $subject;
        $subject .=
          defined( $field->subfield('a') ) ? $field->subfield('a') : '';

#          unless ( 
#            substr( $field->subfield('a'), -1 ) eq ')' 
#            ||
#            substr( $field->subfield('a'), -1 ) eq '.'               
#          ) {
#            $subject .= ". ";
#          }
        
        if ( defined( $field->subfield('b') ) ) {
          my @subfields = $field->subfield('b');
          foreach my $s ( @subfields ) {
            $subject .= $s . ' ';
          }
        }
            
#        $subject .=
#          defined( $field->subfield('b') ) ? " -- " . $field->subfield('b') : '';

        $subject .= getSubfields( $field, 'n' );

        if ( defined( $field->subfield('x') ) ) {
          my @subfields = $field->subfield('x');
          foreach my $s ( @subfields ) {
            $subject .= " -- " . $s;
          }
        }

        $subject .=
          defined( $field->subfield('t') ) ? " " . $field->subfield('t') : '';
        $subject .=
          defined( $field->subfield('z') )
          ? " -- " . $field->subfield('z')
          : '';
        $subject .=
          defined( $field->subfield('y') )
          ? " -- " . $field->subfield('y')
          : '';
        $subject .=
          defined( $field->subfield('v') )
          ? " -- " . $field->subfield('v')
          : '';
        $subject .= ' ';
        push( @subjects, $subject );
    }

    ## 630 Subject Added Entry -- Uniform Title
    my @fields = $_marc->field('630');
    my $firstField = 1;
    foreach my $field (@fields) {
        my $subject;
        
        $subject .=  defined( $field->subfield('a') ) ? $field->subfield('a') : '';
        
        my $n;
        if ( defined( $field->subfield('n') ) ) {
          $n = 1;
          my @subfields = $field->subfield('n');
          foreach my $s ( @subfields ) {
            $subject .= '. ' . $s;
          }
        }
            
        if ( defined( $field->subfield('p') ) ) {
          my @subfields = $field->subfield('p');
          foreach my $s ( @subfields ) {
            if ( $n ) {
              $subject .= ', ';
              $n = 0;
            } else {
              $subject .= '. ';
            }
            $subject .= $s . ' ';
          }
        }

        $subject .=  defined( $field->subfield('l') ) ? '. ' . $field->subfield('l') : '';

        if ( defined( $field->subfield('k') ) ) {
          my @subfields = $field->subfield('k');
          foreach my $s ( @subfields ) {
            $subject .= '. ' . $s;
          }
        }

        $subject .=  defined( $field->subfield('s') ) ? '. ' . $field->subfield('s') : '';

        $subject .=  defined( $field->subfield('f') ) ? '. ' . $field->subfield('f') : '';

        if ( defined( $field->subfield('v') ) ) {
          my @subfields = $field->subfield('v');
          foreach my $s ( @subfields ) {
            $subject .= $s . ' ';
          }
        }

        if ( defined( $field->subfield('x') ) ) {
          my @subfields = $field->subfield('x');
          foreach my $s ( @subfields ) {
            $subject .= $s . ' ';
          }
        }
        
        if ( defined( $field->subfield('y') ) ) {
          my @subfields = $field->subfield('y');
          foreach my $s ( @subfields ) {
            $subject .= $s . ' ';
          }
        }
        
        if ( defined( $field->subfield('z') ) ) {
          my @subfields = $field->subfield('z');
          foreach my $s ( @subfields ) {
            $subject .= $s . ' ';
          }
        }

        $subject .= '.';
        
        push( @subjects, $subject );
    }


    ## 650 Subject Added Entry - Topical Term
    @fields = $_marc->field('650');
    foreach my $field (@fields) {
        my $subject;
        $subject .=
          defined( $field->subfield('a') ) ? $field->subfield('a') : '';
        
        if ( defined( $field->subfield('x') ) ) {
          my @subfields = $field->subfield('x');
          foreach my $s ( @subfields ) {
            $subject .= " -- " . $s;
          }
        }

        if ( defined( $field->subfield('z') ) ) {
          my @subfields = $field->subfield('z');
          foreach my $s ( @subfields ) {
            $subject .= " -- " . $s;
          }
        }

        $subject .=
          defined( $field->subfield('y') )
          ? " -- " . $field->subfield('y')
          : '';

        if ( defined( $field->subfield('v') ) ) {
          my @subfields = $field->subfield('v');
          foreach my $s ( @subfields ) {
            $subject .= " -- " . $s;
          }
        }

        $subject .= ' ';

        push( @subjects, $subject );
    }

    ## 651 Subject Added Entry - Geographic Name
    @fields = $_marc->field('651');
    foreach my $field (@fields) {
        my $subject;
        $subject .=
          defined( $field->subfield('a') ) ? $field->subfield('a') : '';
          
        if ( defined( $field->subfield('x') ) ) {
          my @subfields = $field->subfield('x');
          foreach my $s ( @subfields ) {
            $subject .= " -- " . $s;
          }
        }
        $subject .=
          defined( $field->subfield('z') )
          ? " -- " . $field->subfield('z')
          : '';
        $subject .=
          defined( $field->subfield('y') )
          ? " -- " . $field->subfield('y')
          : '';
        $subject .=
          defined( $field->subfield('v') )
          ? " -- " . $field->subfield('v')
          : '';
        $subject .= ' ';
        push( @subjects, $subject );
    }

    return @subjects;
}

sub getAuthors {
    my ($item) = @_;

    my @authors;

    ## 700 Added Entry - Personal Name
    my @fields = $_marc->field('700');
    foreach my $field (@fields) {
        my $author;
        $author .=
          defined( $field->subfield('a') ) ? ' ' . $field->subfield('a') : '';
        $author .=
          defined( $field->subfield('q') ) ? ' ' . $field->subfield('q') : '';
        $author .=
          defined( $field->subfield('b') ) ? ' ' . $field->subfield('b') : '';
        $author .=
          defined( $field->subfield('c') ) ? ' ' . $field->subfield('c') : '';
        $author .=
          defined( $field->subfield('d') ) ? ' ' . $field->subfield('d') : '';
        $author .=
          defined( $field->subfield('e') ) ? ' ' . $field->subfield('e') : '';
        $author .=
          defined( $field->subfield('t') ) ? ' ' . $field->subfield('t') : '';
        push( @authors, $author );
    }

    ## 710 Added Entry - Corporate Name
    my @fields = $_marc->field('710');
    foreach my $field (@fields) {
        my $author;
        $author .=
          defined( $field->subfield('a') ) ? ' ' . $field->subfield('a') : '';
        if ( defined( $field->subfield('b') ) ) {
          my @subfields = $field->subfield('b');
          foreach my $s ( @subfields ) {
            $author .= " -- " . $s;
          }
        }        
        if ( defined( $field->subfield('e') ) ) {
          my @subfields = $field->subfield('e');
          foreach my $s ( @subfields ) {
            $author .= " -- " . $s;
          }
        }        
        $author .=
          defined( $field->subfield('t') ) ? ' ' . $field->subfield('t') : '';
        push( @authors, $author );
    }

    ## 730 Added Entry - Uniform Title
    my @fields = $_marc->field('730');
    foreach my $field (@fields) {
        my $author;
        $author .=
          defined( $field->subfield('a') ) ? ' ' . $field->subfield('a') : '';

        my $n;
        if ( defined( $field->subfield('n') ) ) {
          $n = 1;
          my @subfields = $field->subfield('n');
          foreach my $s ( @subfields ) {
            $author .= ". " . $s;
          }
        }        

        if ( defined( $field->subfield('p') ) ) {
          my @subfields = $field->subfield('p');
          foreach my $s ( @subfields ) {
            if ( $n ) { $author .= ', '; $n = 0; } else { $author .= '. '; }
            $author .= $s;
          }
        }    
            
        $author .=
          defined( $field->subfield('l') ) ? '. ' . $field->subfield('l') : '';

        if ( defined( $field->subfield('k') ) ) {
          my @subfields = $field->subfield('k');
          foreach my $s ( @subfields ) {
            $author .= ". " . $s;
          }
        }    

        $author .=
          defined( $field->subfield('s') ) ? '. ' . $field->subfield('s') : '';

        $author .=
          defined( $field->subfield('f') ) ? '. ' . $field->subfield('f') : '';

        $author .= '.';
        
        push( @authors, $author );
    }
    

    ## 740 Added Entry - Uncontrolled Related / Analytical Title
    my @fields = $_marc->field('740');
    foreach my $field (@fields) {
        my $author;
        $author .=
          defined( $field->subfield('a') ) ? ' ' . $field->subfield('a') : '';
        $author .=
          defined( $field->subfield('n') ) ? ' ' . $field->subfield('n') : '';
        $author .=
          defined( $field->subfield('p') ) ? ' ' . $field->subfield('p') : '';
        $author .=
          defined( $field->subfield('h') ) ? ' ' . $field->subfield('h') : '';
        push( @authors, $author );
    }

    return @authors;
}

sub getTitle {
    my ( $ignoreSubtitle ) = @_;

    my $titleLine;
    if ( defined( $_marc->field('245') ) && defined( $_marc->field('245')->subfield('a') ) ) {
      $titleLine  = $_marc->field('245')->subfield('a');
    }

    if ( !$ignoreSubtitle && defined( $_marc->field('245') ) && defined( $_marc->field('245')->subfield('h') ) ) {
        $titleLine .= ' ' . $_marc->field('245')->subfield('h');
    }    
    
    if ( !$ignoreSubtitle && defined( $_marc->field('245') ) && defined( $_marc->field('245')->subfield('b') ) ) {
        $titleLine .= ' ' . $_marc->field('245')->subfield('b');
    }

    chomp($titleLine);
    if ( $titleLine =~ /\/$/ ) {    ## If line ends with a '/'
        chop($titleLine);
    }
    if ( $titleLine =~ /\.$/ ) {    ## If line ends with a '/'
        chop($titleLine);
    }

    return $titleLine;
}

sub getAuthor {
    my $authorLine;
    
    if ( $_marc->field('100') ) {
      $authorLine = $_marc->field('100')->subfield('a');
      
      if ( $_marc->field('100')->subfield('c') ) {
          $authorLine .= ' ' . $_marc->field('100')->subfield('c');
      }
      
      if ( $_marc->field('100')->subfield('q') ) {
          $authorLine .= ' (' . $_marc->field('100')->subfield('q') . ')';
          ## If subfield d is present, subfield q should end with a comma
          if ( $_marc->field('100')->subfield('d') ) {
            $authorLine .= ',';
          }
      }
      
      my @subfields = $_marc->field('100')->subfield('b');
      foreach my $s ( @subfields ) {
        $authorLine .= ' ' . $s;
      }
      
      if ( $_marc->field('100')->subfield('d') ) {
          $authorLine .= ' ' . $_marc->field('100')->subfield('d');
      }
      
    } elsif ( $_marc->field('110') ) {
      $authorLine = $_marc->field('110')->subfield('a');
      if ( $_marc->field('110')->subfield('b') ) {
          my @subfields = $_marc->field('110')->subfield('b');
          foreach my $s ( @subfields ) {
            $authorLine .= '. ' . $s;
          }
      }
      if ( $_marc->field('110')->subfield('t') ) {
          $authorLine .= '. ' . $_marc->field('100')->subfield('c');
      }
    } elsif ( $_marc->field('130') ) {
      $authorLine = $_marc->field('130')->subfield('a');

      if ( $_marc->field('130')->subfield('n') ) {
          my @subfields = $_marc->field('130')->subfield('n');
          foreach my $s ( @subfields ) {
            $authorLine .= '. ' . $s;
          }
      }

      if ( $_marc->field('130')->subfield('p') ) {
          my @subfields = $_marc->field('130')->subfield('p');
          foreach my $s ( @subfields ) {
            $authorLine .= '. ' . $s;
          }
      }

      if ( $_marc->field('130')->subfield('l') ) {
          $authorLine .= '. ' . $_marc->field('130')->subfield('l');
      }

      if ( $_marc->field('130')->subfield('k') ) {
          my @subfields = $_marc->field('130')->subfield('k');
          foreach my $s ( @subfields ) {
            $authorLine .= '. ' . $s;
          }
      }

      if ( $_marc->field('130')->subfield('s') ) {
          $authorLine .= '. ' . $_marc->field('130')->subfield('s');
      }

      if ( $_marc->field('130')->subfield('f') ) {
          $authorLine .= '. ' . $_marc->field('130')->subfield('f');
      }
      
    }
    
    return $authorLine;
}

sub formatTracings {
    my ( @tracings ) = @_;

    my @formattedTracings;

    my $indent = INDENT;
    
    my $textLine;
    while ( my $tracing = shift( @tracings ) ) {
      if ( length( $textLine ) < MAX_STRING_WIDTH - length( $tracing ) - $indent ) {
        $textLine .= $tracing . ' ';
      } else {
        push ( @formattedTracings, $textLine );
        $textLine = $tracing;
      }
      $indent = 0;      
    }
    push ( @formattedTracings, $textLine );


    return ( @formattedTracings );
}

sub printCardNumber {
  my ( $cardNumber ) = @_;
  prText( 400, Y_TOP - Y_OFFSET * ( $cardNumber - 1 ) - FH, "Card $_currentCard_number" );
}

sub getFields {
  my ( $field, $subfield, $precededBy, $postcededBy, $separator ) = @_;
  if ( ! $separator ) { $separator = ' '; }

  my $field = $_marc->field( $field );
  
  if ( $field ) {
    my @subfields = $field->subfield( $subfield );

    my $line = pop( @subfields );

    foreach my $sf ( @subfields ) {
      $line .= $separator . $sf;
    }
  
    $line = $precededBy . $line . $postcededBy;
    return $line;
  }
}

## Must be used instead of getFields for repeatable fields.
## $field is a $marc->field( '234' ) object
## $subfield is a string
sub getSubfields {
  my $string;

  my ( $field, $subfield, $precededBy, $postcededBy, $separator ) = @_;
  if ( defined( $field->subfield( $subfield ) ) ) {
    my @subfields = $field->subfield( $subfield );
    foreach my $s ( @subfields ) {
      $string .= $separator . $s;
    }
  }
  
  return $precededBy . $string . $postcededBy;
}


# Perl trim function to remove whitespace from the start and end of the string
sub trim($) {
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
}
